package goblincave.gitea.nes;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.RandomAccessFile;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.LinkedHashMap;
import java.util.Set;

import javaFlacEncoder.FLAC_FileEncoder;

import org.apache.commons.io.FilenameUtils;

import com.diogonunes.jcolor.Ansi;
import com.diogonunes.jcolor.Attribute;

/**
 * Ridge Racer 6 Audio Extractor
 * @author Nes
 * @version 1.0, 2024-07-14
 */
public class AudioExtractor {
	
	/**
	 * ANSI code to go up one line, followed by a carriage return.
	 * Useful for overwriting user input
	 */
	private final static String RETURN_UP = "\033[F\r";
	private final static String PALE_YELLOW = "\u001B[93m";
	
	/**
	 * BIN files must first be extracted from game disk to use this program.
	 * By default uses program directory for I/O and saves audio as WAV.<br>
	 * 
	 * You can simply run this program in the same directory as <b>pack_*.bin</b> files
	 * to launch with default arguments: <pre>extract currentDir currentDir wav</pre>
	 * 
	 * @param args - 4 optional arguments<ol>
	 * <li>"extract" or "pack" mode</li>
	 * <li>package directory containing <b>pack_*.bin</b> files</li>
	 * <li>extract directory for unpacked audio files</li>
	 * <li>extract BGM in "WAV" or "FLAC" format</li></ol>
	 * @throws URISyntaxException 
	 */
	public static void main(String[] args) throws URISyntaxException {
		
		System.out.println(Ansi.colorize(logoFormat("Ridge Racer 6"), Attribute.BRIGHT_RED_TEXT()));
		System.out.println(Ansi.colorize(logoFormat("Audio Extractor", true, 8), Attribute.YELLOW_TEXT()));
		System.out.println("Program:\tRidge Racer 6 Audio Extractor " + Ansi.colorize("v1.0", Attribute.BRIGHT_GREEN_TEXT()) + " by " + Ansi.colorize("Nes", Attribute.TEXT_COLOR(252, 42, 124)));
		System.out.println("Repository:\t" + Ansi.colorize("https://gitea.goblincave.synology.me/Nes/RR6AudioExtractor", Attribute.CYAN_TEXT()));
		
		if(args.length > 0) {
			System.out.println("Arguments\t:");
			for(int i = 0; i < args.length; i++)
				System.out.print(Ansi.colorize(args[i] + ' ', Attribute.BRIGHT_GREEN_TEXT()));
		}
		
		if(args.length == 0) { // no arguments provided
			
			// determine which operation to execute
			
			BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
			String prompt = "\n" + Ansi.colorize("Operations", Attribute.BRIGHT_YELLOW_TEXT(), Attribute.BOLD()) + ":\n"
					+ "- " + Ansi.colorize("Extract", Attribute.BRIGHT_RED_TEXT()) + " audio tracks from package\n"
					+ "- " + Ansi.colorize("Package", Attribute.YELLOW_TEXT()) + " audio tracks into package\n"
					+ "- " + Ansi.colorize("Patch", Attribute.BRIGHT_GREEN_TEXT()) + " audio track to loop indefinitely\n"
					+ "- " + Ansi.colorize("Print", Attribute.BRIGHT_BLUE_TEXT()) + " Ridge Racer 6 logo style ASCII text\n"
					+ "- " + Ansi.colorize("Exit", Attribute.BRIGHT_MAGENTA_TEXT()) + " program\n"
					+ "\nPlease select an operation:" + PALE_YELLOW + " ";
			String input;
			{
				String[] values = {"extract", "package", "patch", "print", "exit"};
				input = retrieveInput(reader, prompt, values);
			}
			
			int operation = -1;
			switch(input.toLowerCase()) {
				case "extract":
					operation = 0;
					System.out.println(RETURN_UP + Ansi.colorize("Please select an operation: ") + Ansi.colorize(input, Attribute.BRIGHT_RED_TEXT()));
					break;
				case "pack":
				case "package":
					operation = 1;
					System.out.println(RETURN_UP + Ansi.colorize("Please select an operation: ") + Ansi.colorize(input, Attribute.YELLOW_TEXT()));
					break;
				case "patch":
					operation = 2;
					System.out.println(RETURN_UP + Ansi.colorize("Please select an operation: ") + Ansi.colorize(input, Attribute.BRIGHT_GREEN_TEXT()));
					break;
				case "logo":
				case "print":
					operation = 3;
					System.out.println(RETURN_UP + Ansi.colorize("Please select an operation: ") + Ansi.colorize(input, Attribute.BRIGHT_BLUE_TEXT()));
					break;
				case "quit":
				case "exit":
					operation = 4;
					System.out.println(RETURN_UP + Ansi.colorize("Please select an operation: ") + Ansi.colorize(input, Attribute.BRIGHT_MAGENTA_TEXT()));
					break;
			}
			
			switch(operation) {
				case 0: // extract
					// Determine package directory
					File currentDir = new File(new File(AudioExtractor.class.getProtectionDomain().getCodeSource().getLocation().toURI()).getParent());
					boolean packDirFound = false;
					do {
						System.out.println(Ansi.colorize("\nPackage Directory:", Attribute.BOLD(), Attribute.CYAN_TEXT()));
						System.out.println("- " + Ansi.colorize("Current", Attribute.BRIGHT_RED_TEXT()) + " directory:\t" + Ansi.colorize(currentDir.getPath(), Attribute.BRIGHT_CYAN_TEXT()));
						
					} while(!packDirFound);
					break;
				case 1: // pack
					break;
				case 2: // patch
					break;
				case 3: // print
					prompt = "\nCommand parameters: " + Ansi.colorize("Print", Attribute.BRIGHT_BLUE_TEXT()) + Ansi.colorize(" [color]", Attribute.BRIGHT_YELLOW_TEXT()) + Ansi.colorize(" [message]", Attribute.BRIGHT_RED_TEXT()) 
							+ "\n" 
							+ "\n" + Ansi.colorize("Colors", Attribute.BRIGHT_YELLOW_TEXT(), Attribute.BOLD()) + ":"
							+ "\n- " + Ansi.colorize("Default", Attribute.CLEAR())
							+ "\n- " + Ansi.colorize("Black", Attribute.BLACK_TEXT()) + ", " + Ansi.colorize("Gray", Attribute.BRIGHT_BLACK_TEXT()) + ", " + Ansi.colorize("White", Attribute.BRIGHT_WHITE_TEXT())
							+ "\n- " + Ansi.colorize("Indigo", Attribute.TEXT_COLOR(0x7D, 0x1A, 0xFF)) + ", " + Ansi.colorize("Purple", Attribute.BRIGHT_MAGENTA_TEXT()) + ", " + Ansi.colorize("Violet", Attribute.TEXT_COLOR(0xEE, 0x82, 0xEE))
							+ "\n- " + Ansi.colorize("Red", Attribute.RED_TEXT()) + ", " + Ansi.colorize("Cinnabar", Attribute.BRIGHT_RED_TEXT()) + ", " + Ansi.colorize("Pink", Attribute.TEXT_COLOR(0xFF, 0xC0, 0xCB))
							+ "\n- " + Ansi.colorize("Orange", Attribute.TEXT_COLOR(0xFF, 0x68, 0x1F)) + ", " + Ansi.colorize("Gold", Attribute.YELLOW_TEXT()) + ", " + Ansi.colorize("Yellow", Attribute.TEXT_COLOR(0xFF, 0xEA, 0x00))
							+ "\n- " + Ansi.colorize("Green", Attribute.BRIGHT_GREEN_TEXT()) + ", " + Ansi.colorize("Chartreuse", Attribute.TEXT_COLOR(0xD1, 0xEB, 0x27)) + ", " + Ansi.colorize("Beige", Attribute.BRIGHT_YELLOW_TEXT())
							+ "\n- " + Ansi.colorize("Blue", Attribute.BRIGHT_BLUE_TEXT()) + ", " + Ansi.colorize("Teal", Attribute.TEXT_COLOR(0x00, 0x93, 0xCF)) + ", " + Ansi.colorize("Cyan", Attribute.BRIGHT_CYAN_TEXT())
							+ "\n" 
							+ "\n" + Ansi.colorize("Please select a color:") + PALE_YELLOW + " ";
					String[] values = {"Default", "Black", "Gray", "White",	
							"Indigo", "Purple", "Violet", "Red", "Cinnabar", "Pink", 
							"Orange", "Gold", "Yellow",	"Green", "Chartreuse", "Beige",	
							"Blue", "Teal", "Cyan"};
					input = retrieveInput(reader, prompt, values);
					Attribute color = null;
					switch(input.toLowerCase()) {
						case "default":		color = Attribute.CLEAR(); break;
						case "black": 		color = Attribute.BLACK_TEXT(); break;
						case "gray": 		color = Attribute.BRIGHT_BLACK_TEXT(); break;
						case "white": 		color = Attribute.BRIGHT_WHITE_TEXT(); break;
						case "indigo": 		color = Attribute.TEXT_COLOR(0x7D, 0x1A, 0xFF); break;
						case "purple": 		color = Attribute.BRIGHT_MAGENTA_TEXT(); break;
						case "violet": 		color = Attribute.TEXT_COLOR(0xEE, 0x82, 0xEE); break;
						case "red": 		color = Attribute.RED_TEXT(); break;
						case "cinnabar": 	color = Attribute.BRIGHT_RED_TEXT(); break;
						case "pink": 		color = Attribute.TEXT_COLOR(0xFF, 0xC0, 0xCB); break;
						case "green": 		color = Attribute.BRIGHT_GREEN_TEXT(); break;
						case "chartreuse": 	color = Attribute.TEXT_COLOR(0xD1, 0xEB, 0x27); break;
						case "beige":		color = Attribute.BRIGHT_YELLOW_TEXT(); break;
						case "blue":		color = Attribute.BRIGHT_BLUE_TEXT(); break;
						case "teal":		color = Attribute.TEXT_COLOR(0x00, 0x93, 0xCF); break;
						case "cyan":		color = Attribute.BRIGHT_CYAN_TEXT(); break;
					}
					System.out.println(RETURN_UP + Ansi.colorize("Please select a color: ") + Ansi.colorize(input, color));
					break;
				case 4: // exit
					System.out.println("Program closed.");
					System.exit(0);
					break;
			}

		}
		System.exit(0);
		
		if(args.length > 0) {
			switch(args[0]) {
				
				case "help":
					if(args.length > 1)
						switch(args[1]) {
							case "extract":
								
								break;
							case "pack":
								break;
							case "print":
								break;
						}
					System.out.println("java -jar RR6AudioExtractor.jar [operation] [package] [extract] [option]");
					break;
					
				case "extract":
					// TODO 
					break;
					
				case "pack":
					// TODO
					break;
					
			}
				
		}
		
		// Confirm input/output mode
		boolean packMode = false;
		if(args.length > 0)
			packMode = args[0].equalsIgnoreCase("pack");
		
		// Confirm package directory, else abort
		File packDirectory;
		if(args.length > 1)
			packDirectory = Path.of(args[1]).toFile();
		else packDirectory = new File(new File(AudioExtractor.class.getProtectionDomain().getCodeSource().getLocation().toURI()).getParent());
		if(packMode && !packDirectory.exists() && !packDirectory.mkdir()) {
			System.out.println("Package output directory " + packDirectory + " does not exist, and could not be created. Program execution aborted.");
			System.exit(1);
		} else if(!packMode && !packDirectory.exists()) {
			System.out.println("Package input directory " + packDirectory + " does not exist. Program execution aborted.");
			System.exit(1);
		}
		
		// Confirm extract directory, else abort
		File extractDirectory;
		if(args.length > 2)
			extractDirectory = Path.of(args[2]).toFile();
		else extractDirectory = packDirectory;
		if(!packMode && !extractDirectory.exists() && !extractDirectory.mkdir()) {
			System.out.println("Extract output directory " + extractDirectory + " does not exist, and could not be created. Program execution aborted.");
			System.exit(2);
		}
		
		// Confirm BGM output format
		boolean compressBGM = false;
		if(args.length > 3)
			compressBGM = args[3].equalsIgnoreCase("FLAC");
		
		// Begin operation
		if(packMode)
			pack(packDirectory, extractDirectory);
		else extract(packDirectory, extractDirectory, compressBGM);
		
	}
	
	/**
	 * Prompts the user for input and validates that the input matches one of the provided values.
	 * 
	 * @param reader
	 * @param prompt
	 * @param values
	 * @return input
	 */
	private static String retrieveInput(BufferedReader reader, String prompt, String[] values) {
		
		String input = null;
		boolean valid = false;
		do {
			System.out.print(prompt);
			try {
				input = reader.readLine();
				for(int i = 0; i < values.length; i++)
					if(input.equalsIgnoreCase(values[i])) {
						valid = true;
						break;
					}
				if(!valid)
					System.out.println(Ansi.colorize("Invalid selection."));
			} catch (IOException e) {
				System.out.println(Ansi.colorize("Invalid selection."));
			}
		} while(!valid);
		
		return input;
		
	}

	private static void pack(File packDirectory, File extractDirectory) {
		
		// Identify extracted folders
		File[] packages = new File(extractDirectory.toString()).listFiles(new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.toLowerCase().startsWith("pack_") && !name.toLowerCase().endsWith(".bin");
			}
		}); 
        
		for(int i = 0; i < packages.length; i++)
			System.out.println(packages[i].getName());
		
		for(File packDir : packages) {
			
			// 1. Compile list of WAV files in extracted directories
			File[] audioTracks = new File(packDir.toString()).listFiles(new FilenameFilter() {
				public boolean accept(File dir, String name) {
					return name.toLowerCase().endsWith(".wav");
				}
			});
			
			// 2. Check if pack files already exist in package directory
			File pack = new File(packDir.getPath() + ".bin");
			if(pack.exists());
			
			// 3. If pack file already exists, insert WAV files into existing file.
			// 4. If pack file doesn't exist, create new file then insert WAV files.
		}
		// Hypothesis 1: Does Ridge Racer 6 use a predetermined address list for reading tracks from package?
		// Test 1: Try changing the address of a track and see if the game can handle it successfully.
		
		// Hypothesis 2: Does Ridge Racer 6 loop tracks in their entirety, or do they use loop points defined in the file header?
		// Test 1: Try editing loop information and see if it affects in-game playback.
		// Test 2: Try replacing a track with a WAV file without loop data and see if it loops in in-game playback.
		
	}

	public static void extract(File packDirectory, File extractDirectory, boolean compressBGM) {
		
		// TODO Delete
		System.out.println("packDirectory:\t" + packDirectory);
		System.out.println("extractDirectory:\t" + extractDirectory);
		
		// Identify target binary files
		File[] packages = new File(packDirectory.toString()).listFiles(new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.toLowerCase().endsWith(".bin");
			}
		});
		
		// Identify and extract audio within binary files
		for(int i = 0; i < packages.length; i++) {
			
			RandomAccessFile source;
			LinkedHashMap<Integer, Integer> tracklist = null;
			try {
				source = new RandomAccessFile(packages[i], "r");
				tracklist = findAudioTracks(source);
			} catch (FileNotFoundException e) {
				System.out.println("Binary file:\t" + packages[i] + "\n was not found. File skipped.");
				continue;
			} catch (IOException e) {
				System.out.println("Binary file:\t" + packages[i] + "\n could not be read. File skipped.");
				continue;
			}
			
			//TODO Delete
			System.out.println("packages[" + i + "]:\t" + packages[i]);
					
			File dir = new File(extractDirectory.getPath() + "/" + FilenameUtils.removeExtension(packages[i].getName()));
			if(!dir.exists())
				dir.mkdir();
			
			// Extract and write WAV files in directory
			Set<Integer> keys = tracklist.keySet();
			int track = 1;
			
			for(Integer key : keys) {
				
				String name = String.format("track%04d", track);
				{ // Extract WAV files
					File wav = null;
					try {
							
						// create file for storing WAV data
						wav = new File(dir.getPath() + "/" + name + ".wav");
						if(!wav.exists())
							wav.createNewFile();
					
						//TODO Delete
						System.out.println("Saving track " + String.format("%d (@0x%08X)", track, key) + " to " + wav + String.format(" (%d bytes)", tracklist.get(key)));
						
						// write selection to file
						try (
								FileInputStream inStream = new FileInputStream(packages[i]); 
								FileOutputStream outStream = new FileOutputStream(wav)
						) {
							inStream.skipNBytes(key.intValue());
							byte[] trackBytes = new byte[tracklist.get(key)];
							int readBytes = inStream.read(trackBytes);
							outStream.write(trackBytes, 0, readBytes);
						} catch (IOException e) {
							e.printStackTrace();
							System.exit(3);
						}
						
					} catch (IOException e) {
						System.out.println("An error occurred when attempting to write " + name + " to file:\t" + wav);
						e.printStackTrace();
						System.exit(3);
					}
				}
				
				track++;
				
			}
			
			if(compressBGM && packages[i].getName().equals("pack_bgm.bin")) {

				// find WAV files
				File[] WAV_Files = dir.listFiles(new FilenameFilter() {
					public boolean accept(File dir, String name) {
						return name.toLowerCase().endsWith(".wav");
					}
				});
				
				// convert WAV files to FLAC
				FLAC_FileEncoder ffe = new FLAC_FileEncoder();
								
				for(File wav : WAV_Files) {
				
					try {
						File flac = new File(dir.getPath() + "/" + FilenameUtils.removeExtension(wav.getName()) + ".flac");
						if(!flac.exists())
							flac.createNewFile();
						
						//TODO Delete
						System.out.println("Compressing WAV to FLAC:\t" + flac);
					
						ffe.encode(wav, flac);
					} catch (IOException e) {
						System.out.println("An error occurred when attempting to write file.");
						e.printStackTrace();
						System.exit(3);
					}
					
				}
				
				System.out.print("Deleting WAV files");
				// delete WAV files
				for(File wav : WAV_Files) {
					boolean deleted = false;
					while(!deleted) {
						try {
							Files.delete(wav.toPath());
							deleted = true;
						} catch(Exception e) {
							System.out.print('.');
						}
					}
				}
				
			}
			
		}
		
	}

	/**
	 * Creates a map of audio tracks within provided file.
	 * 
	 * @param file
	 * @return Tracklist containing address-length pairs
	 * @throws IOException
	 */
	public static LinkedHashMap<Integer, Integer> findAudioTracks(RandomAccessFile file) throws IOException {
		
		LinkedHashMap<Integer, Integer> tracklist = new LinkedHashMap<Integer, Integer>();
		int key = 0x52494646; // "RIFF" in ASCII representation 
		
		// Files start on addresses divisible by 0x800
		for(int a = 0; a + 0x800 < file.length(); a += 0x800) {
			
			file.seek(a);
			
			// File header starts with "RIFF" sequence
			if(file.readInt() == key) {
				
				// Next 4 bytes specify file length beyond first 8 bytes, in little-endian representation
				int length = 8;
				// Calculate remaining length of file
				for(int d = 0; d < 4; d++) {
					file.seek(a + 4 + d);
					length += file.readUnsignedByte() * ((int) Math.pow(16, 2 * d));
				}
				
				// Record file position in map
				tracklist.put(a, length);
			}
			
		}
		
		return tracklist;
		
	}
	
	/**
	 * Creates an ASCII logo in RR6 style using the provided text.
	 * Supports alphanumeric text with spaces only.
	 * 
	 * @param text
	 * @param italicize logo
	 * @return logo formatted text
	 */
	public static String logoFormat(String text) {
		return logoFormat(text, true, 10);
	}
	
	/**
	 * Creates an ASCII logo in RR6 style using the provided text.
	 * Supports alphanumeric text with spaces only.
	 * 
	 * @param text
	 * @param height 
	 * @param italicize logo
	 * @return logo formatted text
	 */
	public static String logoFormat(String text, boolean italic, int height) {
		
		// Initialize logo lines
		String[] logo = new String[10];
		for(int i = 0; i < logo.length; i++)
			logo[i] = new String();
		
		boolean spaced = false;
		for(int i = 0; i < text.length(); i++) {
			char c = text.charAt(i);

			// Add space from previous character
			if(i > 0) {
				if(c != ' ' && c!= ':' && !spaced) {
					for(int line = 0; line < logo.length - 1; line++)
						logo[line] += "   ";
					logo[9] += "███";
				}
				spaced = false;
			}
			
			// Add character to logo
			switch(c) {
				
				// Space
				case ':':
				case ' ':
					for(int line = 0; line < logo.length; line++)
						logo[line] += "    ";
					spaced = true;
					break;
					
				// Numbers
				case '0':
					if(logo[0].length() > 0)
						for(int x = 0; x < logo.length; x++)
							logo[x] = logo[x].substring(0, logo[x].length() - 1);
					logo[0] += " ██████████████████████    ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████        ████████   ";
					logo[3] += "████████        ████████   ";
					logo[4] += "████████        ████████   ";
					logo[5] += "████████        ████████   ";
					logo[6] += "████████        ████████   ";
					logo[7] += "████████        ████████   ";
					logo[8] += "████████████████████████   ";
					logo[9] += " ██████████████████████    ";
					spaced = true;
					break;	
				case '1':
					logo[0] += " ███████████████   ";
					logo[1] += "████████████████   ";
					logo[2] += "        ████████   ";
					logo[3] += "        ████████   ";
					logo[4] += "        ████████   ";
					logo[5] += "        ████████   ";
					logo[6] += "        ████████   ";
					logo[7] += "        ████████   ";
					logo[8] += "        ████████   ";
					logo[9] += "        ████████   ";
					spaced = true;
					break;
				case '2':
					logo[0] += " ███████████████████████   ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████        ████████   ";
					logo[3] += "                ████████   ";
					logo[4] += "  ██████████████████████   ";
					logo[5] += "████████████████████████   ";
					logo[6] += "████████                   ";
					logo[7] += "████████                   ";
					logo[8] += "████████████████████████   ";
					logo[9] += "███████████████████████    ";
					spaced = true;
					break;
				case '3':
					logo[0] += " ██████████████████████    ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████        ████████   ";
					logo[3] += "                ████████   ";
					logo[4] += "        ███████████████    ";
					logo[5] += "        ██████████████     ";
					logo[6] += "                ████████   ";
					logo[7] += "████████        ████████   ";
					logo[8] += "████████████████████████   ";
					logo[9] += " ██████████████████████    ";
					spaced = true;
					break;
				case '4':
					logo[0] += "         █████████████     ";
					logo[1] += "       ███████████████     ";
					logo[2] += "     ████       ██████     ";
					logo[3] += "   ██████       ██████     ";
					logo[4] += " ███████████████████████   ";
					logo[5] += "████████████████████████   ";
					logo[6] += "                ██████     ";
					logo[7] += "                ██████     ";
					logo[8] += "                ██████     ";
					logo[9] += "                ██████     ";
					spaced = true;
					break;
				case '5':
					logo[0] += " ███████████████████████   ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████                   ";
					logo[3] += "████████                   ";
					logo[4] += "████████████████████████   ";
					logo[5] += "████████████████████████   ";
					logo[6] += "                ████████   ";
					logo[7] += "                ████████   ";
					logo[8] += "████████████████████████   ";
					logo[9] += "███████████████████████    ";
					spaced = true;
					break;
				case '6':
					logo[0] += " ███████████████████████   ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████                   ";
					logo[3] += "████████                   ";
					logo[4] += "████████████████████████   ";
					logo[5] += "████████████████████████   ";
					logo[6] += "████████        ████████   ";
					logo[7] += "████████        ████████   ";
					logo[8] += "████████████████████████   ";
					logo[9] += "███████████████████████    ";
					spaced = true;
					break;
				case '7':
					logo[0] += " ███████████████████████   ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████        ████████   ";
					logo[3] += "                ████████   ";
					logo[4] += "                ████████   ";
					logo[5] += "                ████████   ";
					logo[6] += "                ████████   ";
					logo[7] += "                ████████   ";
					logo[8] += "                ████████   ";
					logo[9] += "                ████████   ";
					spaced = true;
					break;
				case '8':
					logo[0] += " ██████████████████████    ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████        ████████   ";
					logo[3] += "████████        ████████   ";
					logo[4] += "  █████████████████████    ";
					logo[5] += " █████████████████████     ";
					logo[6] += "████████        ████████   ";
					logo[7] += "████████        ████████   ";
					logo[8] += "████████████████████████   ";
					logo[9] += " ██████████████████████    ";
					spaced = true;
					break;
				case '9':
					logo[0] += " ██████████████████████   ";
					logo[1] += "████████████████████████   ";
					logo[2] += "████████        ████████   ";
					logo[3] += "████████        ████████   ";
					logo[4] += " ███████████████████████    ";
					logo[5] += "   █████████████████████     ";
					logo[6] += "                ████████   ";
					logo[7] += "                ████████   ";
					logo[8] += "                ████████   ";
					logo[9] += "                ████████    ";
					spaced = true;
					break;
				
				// Capital letters
				case 'A':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████████████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'B':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "███████████ ";
					logo[6] += "████    ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'C':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████        ";
					logo[4] += "████        ";
					logo[5] += "████        ";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'D':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'E':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████        ";
					logo[4] += "████        ";
					logo[5] += "████████████";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'F':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████        ";
					logo[4] += "████        ";
					logo[5] += "████████████";
					logo[6] += "████        ";
					logo[7] += "████        ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'G':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████        ";
					logo[4] += "████        ";
					logo[5] += "████  ██████";
					logo[6] += "████    ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'H':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████    ████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████████████";
					logo[6] += "████    ████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'I':
					logo[0] += "    ";
					logo[1] += "    ";
					logo[2] += "████";
					logo[3] += "████";
					logo[4] += "████";
					logo[5] += "████";
					logo[6] += "████";
					logo[7] += "████";
					logo[8] += "    ";
					logo[9] += "████";
					break;
				case 'J':
					logo[0] += "        ";
					logo[1] += "        ";
					logo[2] += "  ██████";
					logo[3] += "    ████";
					logo[4] += "    ████";
					logo[5] += "    ████";
					logo[6] += "    ████";
					logo[7] += "████████";
					logo[8] += "        ";
					logo[9] += "████████";
					break;
				case 'K':
					logo[0] += "             ";
					logo[1] += "             ";
					logo[2] += "████    ████ ";
					logo[3] += "████   ████  ";
					logo[4] += "████  ████   ";
					logo[5] += "████ ████    ";
					logo[6] += "████████     ";
					logo[7] += "████  █████  ";
					logo[8] += "        █████";
					logo[9] += "███████   ███";
					spaced = true;
					break;
				case 'L':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████        ";
					logo[3] += "████        ";
					logo[4] += "████        ";
					logo[5] += "████        ";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'M':
					logo[0] += "              ";
					logo[1] += "              ";
					logo[2] += "████      ████";
					logo[3] += "██████  ██████";
					logo[4] += "████ ████ ████";
					logo[5] += "████      ████";
					logo[6] += "████      ████";
					logo[7] += "████      ████";
					logo[8] += "              ";
					logo[9] += "██████████████";
					break;
				case 'N':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████    ████";
					logo[3] += "██████  ████";
					logo[4] += "███████ ████";
					logo[5] += "████ ███████";
					logo[6] += "████  ██████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'O':
					if(logo[0].length() > 0)
						for(int x = 0; x < logo.length; x++)
							logo[x] = logo[x].substring(0, logo[x].length() - 1);
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += " ██████████ ";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += " ██████████ ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'P':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████████████";
					logo[7] += "████        ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'Q':
					logo[0] += "               ";
					logo[1] += "               ";
					logo[2] += " ███████████   ";
					logo[3] += "████    ████   ";
					logo[4] += "████    ████   ";
					logo[5] += "████    ████   ";
					logo[6] += "████    ████   ";
					logo[7] += "████████████   ";
					logo[8] += "          ████ ";
					logo[9] += "█████████   ███";
					spaced = true;
					break;
				case 'R':
					logo[0] += "               ";
					logo[1] += "               ";
					logo[2] += "████████████   ";
					logo[3] += "████    ████   ";
					logo[4] += "████    ████   ";
					logo[5] += "████    ████   ";
					logo[6] += "███████████    ";
					logo[7] += "█████   █████  ";
					logo[8] += "          █████";
					logo[9] += "█████████   ███";
					spaced = true;
					break;
				case 'S':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "████    ████";
					logo[4] += "  ████      ";
					logo[5] += "      ████  ";
					logo[6] += "████    ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'T':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "    ████    ";
					logo[4] += "    ████    ";
					logo[5] += "    ████    ";
					logo[6] += "    ████    ";
					logo[7] += "    ████    ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'U':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████    ████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'V':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████    ████";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "  ████████  ";
					logo[7] += "    ████    ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'W':
					logo[0] += "              ";
					logo[1] += "              ";
					logo[2] += "████      ████";
					logo[3] += "████      ████";
					logo[4] += "████      ████";
					logo[5] += "████ ████ ████";
					logo[6] += "██████  ██████";
					logo[7] += "████      ████";
					logo[8] += "              ";
					logo[9] += "██████████████";
					break;
				case 'X':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████    ████";
					logo[3] += "████    ████";
					logo[4] += " ████  ████ ";
					logo[5] += "    ████    ";
					logo[6] += " ████  ████ ";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'Y':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████    ████";
					logo[3] += "████    ████";
					logo[4] += " ████  ████ ";
					logo[5] += "    ████    ";
					logo[6] += "    ████    ";
					logo[7] += "    ████    ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'Z':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "████████████";
					logo[3] += "        ████";
					logo[4] += "      ████  ";
					logo[5] += "  ████      ";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
					
				// Lowercase letters
				case 'a':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ███████████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████████████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'b':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "████    ████";
					logo[5] += "██████████  ";
					logo[6] += "████    ████";
					logo[7] += "███████████ ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'c':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ███████████";
					logo[4] += "████        ";
					logo[5] += "████        ";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'd':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += "███████████ ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'e':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ███████████";
					logo[4] += "████        ";
					logo[5] += "████████████";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'f':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ███████████";
					logo[4] += "████        ";
					logo[5] += "████████████";
					logo[6] += "████        ";
					logo[7] += "████        ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'g':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ███████████";
					logo[4] += "████        ";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += "███████████ ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'h':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████████████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'i':
					logo[0] += "    ";
					logo[1] += "    ";
					logo[2] += "    ";
					logo[3] += "████";
					logo[4] += "████";
					logo[5] += "████";
					logo[6] += "████";
					logo[7] += "████";
					logo[8] += "    ";
					logo[9] += "████";
					break;
				case 'j':
					logo[0] += "        ";
					logo[1] += "        ";
					logo[2] += "        ";
					logo[3] += "  ██████";
					logo[4] += "    ████";
					logo[5] += "    ████";
					logo[6] += "    ████";
					logo[7] += "████████";
					logo[8] += "    ";
					logo[9] += "████████";
					break;
				case 'k':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "████  ████  ";
					logo[5] += "████████    ";
					logo[6] += "████  ████  ";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'l':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████        ";
					logo[4] += "████        ";
					logo[5] += "████        ";
					logo[6] += "████        ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'm':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████   █████";
					logo[4] += "█████ ██████";
					logo[5] += "████ ██ ████";
					logo[6] += "████    ████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'n':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "█████   ████";
					logo[5] += "████ ██ ████";
					logo[6] += "████   █████";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'o':
					if(logo[0].length() > 0)
						for(int x = 0; x < logo.length; x++)
							logo[x] = logo[x].substring(0, logo[x].length() - 1);
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ██████████ ";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += " ██████████ ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'p':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "██████████  ";
					logo[7] += "████        ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'q':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += " ███████████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "  ██████████";
					logo[7] += "        ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'r':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "██████████  ";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 's':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "████        ";
					logo[5] += "    ████    ";
					logo[6] += "        ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 't':
					if(logo[0].length() > 0)
						for(int x = 0; x < logo.length; x++)
							logo[x] = logo[x].substring(0, logo[x].length() - 1);
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "    ████    ";
					logo[5] += "    ████    ";
					logo[6] += "    ████    ";
					logo[7] += "    ████    ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'u':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "████    ████";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'v':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "████    ████";
					logo[5] += "████    ████";
					logo[6] += "  ████  ████";
					logo[7] += "     █████  ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'w':
					logo[0] += "               ";
					logo[1] += "               ";
					logo[2] += "               ";
					logo[3] += "███   ███   ███";
					logo[4] += "███   ███   ███";
					logo[5] += "███   ███   ███";
					logo[6] += "  ██  ████  ███";
					logo[7] += "    ████  ████ ";
					logo[8] += "               ";
					logo[9] += "███████████████";
					break;
				case 'x':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "  ████ ████ ";
					logo[5] += "    ████    ";
					logo[6] += " ████ ████  ";
					logo[7] += "████    ████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'y':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████    ████";
					logo[4] += "  ████ ████ ";
					logo[5] += "    ████    ";
					logo[6] += "    ████    ";
					logo[7] += "    ████    ";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				case 'z':
					logo[0] += "            ";
					logo[1] += "            ";
					logo[2] += "            ";
					logo[3] += "████████████";
					logo[4] += "      ████  ";
					logo[5] += "    ████    ";
					logo[6] += "  ████      ";
					logo[7] += "████████████";
					logo[8] += "            ";
					logo[9] += "████████████";
					break;
				default:
					System.out.println("Unable to convert character '" + c + "'");
					break;
			}
		}
		
		// Italicize logo
		if(italic) {
			for(int i = 0; i < logo.length; i++) {
				// i = 0 -> 9
				for(int x = logo.length - (i + 1); x > 0; x--)
					// x = 10 - (i + 1)  
					logo[i] = " " + logo[i]; // line 0 gets 10 spaces, line 1 gets 9 spaces... line 9 gets 0 spaces
			}
		}
		
		// Compile logo into single text element
		StringBuilder logoText = new StringBuilder();
		for(int i = logo.length - height; i < logo.length; i++) {
			logoText.append(logo[i] + "\n");
		}
		
		return logoText.toString();
		
	}
	
}
